import uuid
import subprocess
import random
import template_utils   
import haikunator
import sys
import copy
import json
import sample_markov
from midiutil.MidiFile import MIDIFile

import generator

def get_bounds(val, par_a, par_b, par_c, val_mul_a, val_mul_b):
    bounds = [int(par_a * par_b/ (par_b - val*val_mul_a)), int(par_c * par_b / val * val_mul_b)]
    bounds.sort()
    return bounds 
    

def organize_blocks(blocks):
    for i in range(len(blocks)):
        if i: 
            blocks[i]['play_at'] = [blocks[i-1]['play_at'][-1] + blocks[i-1]['number_of_bars'] * blocks[i-1]['number_of_beats_per_bar']]
    return blocks

def shuffle_segments(segments, segment_number_range = range(1, 3)):
    shuffled_segments = []
    for segment in segments: 
        for i in segment_number_range: 
            shuffled_segments.append(copy.deepcopy(segment))
    random.shuffle(shuffled_segments)
    organized = organize_blocks(shuffled_segments)
    return organized


def shuffle_play_at(segments, segment_number_range = range(1, 3)):
    segments_with_repeats = [[segment, random.choice(segment_number_range)] for segment in segments]
    prev_segment = None
    current_play_at = 0
    while segments_with_repeats:
        segment = random.choice(segments_with_repeats)
        while segment == prev_segment and len(segments_with_repeats) != 1: 
            segment = random.choice(segments_with_repeats)

        prev_segment = segment
        i = segments_with_repeats.index(segment)
        segments_with_repeats[i][1] -= 1
        if segments_with_repeats[i][1] == 0: del segments_with_repeats[i]
        segment = segment[0]

        i = segments.index(segment)
        segment_play_at = [current_play_at + j * segments[i]['number_of_bars'] * segments[i]['number_of_beats_per_bar'] for j in range(segments[i]['number_repeats'])]
        segments[i]['play_at'] += segment_play_at
        current_play_at = segments[i]['play_at'][-1] + segments[i]['number_of_bars'] * segments[i]['number_of_beats_per_bar']
    return segments
    

def create_segment_percussion(segment, kwargs):
    percussion = template_utils.create_percussion(segment, no_hits = 2, no_cymbals = random.randint(1, 3))
#    percussion['blocks'][0]['pattern'] = [1] * (segment['number_of_beats_per_bar']%2) + [4]*int(segment['number_of_beats_per_bar']/4)
#    percussion['blocks'][0]['default_accent'] = 80
#    print 'Default pattern is : ', percussion['blocks'][0]['pattern']
    percussion['markov_values'] = False

    for block in percussion['blocks']: 
        pattern = template_utils.generate_pattern(segment, kwargs.get('pattern_percussion_min_len_range', 1) , kwargs.get('pattern_percussion_max_len_range', 4))
        block['pattern'] = pattern
        block['bias_same_note'] = random.choice(kwargs.get('percussion_bias_same_note', range(30, 90, 5)))
        block['default_accent'] = segment['default_accent'] + random.randint(-5, 5) + kwargs.get('percussion_accent_offset', 20)
    return percussion

def get_instrument(program_number, instruments):
    instrument = [x for x in instruments if x['program_number'] == program_number] or[None]
    instrument = instrument[0]
    if not instrument: 
        instrument = {
            'program_number' : program_number, 
        }

    return instrument

def get_random_instrument(instruments, instruments_range):
    reroll = False
    while not reroll: 
        program_number = random.choice(instruments_range)
        instrument = [x for x in instruments if x['program_number'] == program_number] or[{}]
        instrument = instrument[0]
        reroll = random.random() >  instrument.get('reroll_chance', 0)
        if not instrument: 
            instrument = {
                'program_number' : program_number, 
            }

    return instrument


def create_segment(segments, i, kwargs, song_instruments, song_chord, markov_values, instruments = {}):
    song_root_note = random.choice(template_utils.base_notes)
    segment = segments[i]

    prev = {}
    if i: 
        prev = segments[i-1]
        if random.random() < kwargs.get('bpm_change_prob', 0.2): 
            segment['bpm'] = random.choice(kwargs.get('bpm_range', range(150, 540, 15)))
        else:
            segment['bpm'] = prev['bpm']
#        lower_bpm_bound, higher_bpm_bound = kwargs.get('bpm_bounds', get_bounds(prev['bpm'], -5, 540, 25, 0.99, 0.7))
        
#        segment['bpm'] = prev['bpm'] + random.randint(lower_bpm_bound, higher_bpm_bound)
        lower_accent_bound, higher_accent_bound = kwargs.get('default_accent_bounds', get_bounds(prev['default_accent'], -3, 100, 4, 0.9, 0.9))
        
        segment['default_accent'] = min(prev['default_accent'] + random.randint(lower_accent_bound, higher_accent_bound), kwargs.get('max_base_default_accent', 100))
    else: 
        segment['default_accent'] = random.choice(kwargs.get('default_accent_range', range(65, 85)))

    number_segment_instruments_range = kwargs.get('no_segment_instruments_range', range(1, 5 - len(song_instruments)))
    number_segment_instruments = random.choice(number_segment_instruments_range)
    
    segment['number_repeats'] = random.choice(kwargs.get('number_repeats_range', [1]))

    segment_instruments_range = kwargs.get('segment_instruments_range', range(0, 90))
#    segment_instruments = song_instruments + [random.choice(segment_instruments_range) for i in range(number_segment_instruments)]

    segment_instruments = song_instruments + [get_random_instrument(instruments, segment_instruments_range) for i in range(number_segment_instruments)]
#    segment_instruments = [get_instrument(segment_instrument, instruments) for segment_instrument in segment_instruments]
    print 'Instruments for ', segment['track'], ' are : ', segment_instruments, ' and should have : ', number_segment_instruments

    chords = template_utils.generate_random_chord_progression(song_root_note, song_chord,  random.choice(kwargs.get('chords_per_segment_range', range(3, 8))), scale_choices = ['major', 'minor', 'blues'], markov_values = markov_values, exp_var = kwargs.get('chord_exp_var', 5))
    segment['number_of_bars'] = len(chords) * random.choice(kwargs.get('number_segment_bars_range', range(1, 3)))
    
    for instrument in range(len(segment_instruments)):
        #TODO args for low_end and high_end
        low_end = segment_instruments[instrument].get('low_end', random.randint(1, 3))
        high_end = segment_instruments[instrument].get('high_end', min(low_end + random.randint(1, 3), 4))
        print 'My low end is : ', low_end
        print 'My high end is : ', high_end
        
        segment_channel = instrument + 1 #+ segment['track']*len(segment_instruments) + 1 
        print 'Segment channel is : ', segment_channel
        if segment_channel == 10: segment_channel = instrument + len(segments)*len(segment_instruments) + 1 

        pattern = None

        instrument_kwargs = {
            'low_end' : song_root_note + str(low_end), 
            'high_end' : song_root_note + str(high_end), 
            'channel' : segment_channel , 
            'program_number' : segment_instruments[instrument]['program_number'], 
            'default_accent' : segment_instruments[instrument].get('volume', segment['default_accent']) + random.choice(kwargs.get('block_default_accent_range', range(-5, 5))), 
            'bias_same_note' : random.choice(kwargs.get('block_same_note_range', range(10, 100, 15))), 
        }
        max_note_len = kwargs.get('pattern_note_max_len_range', 4)
        min_note_len = kwargs.get('pattern_note_min_len_range', 1)

        if random.random() < kwargs.get('segment_instrument_pattern_chance', 0.5): 
            instrument_kwargs['pattern']= template_utils.generate_pattern(segment, min_note_len, max_note_len)
                     
        segment['blocks'] += template_utils.create_chord_progression(segment, chords = chords, extra_kwargs = instrument_kwargs)


    do_percussion = random.random()
    if not (do_percussion >= kwargs.get('segment_percussion_chance', 0.85) and len(segment['blocks'])/len(segment_instruments) < kwargs.get('skip_percussion_treshold', 3)): 
        percussion = create_segment_percussion(segment, kwargs)           
        segment['blocks'].append(percussion)

    return segment

def generate_song(**kwargs):
    song_root_note = random.choice(template_utils.base_notes)
    song_chord = random.choice(kwargs.get('song_scale', ['major', 'minor']))
    number_of_segments = random.choice(kwargs.get('number_of_segments_range', random.randint(3, 7)))
    bpm_range = kwargs.get('bpm_range', range(150, 540, 15))

    instruments_file = kwargs.get('instruments_file')
    if not instruments_file: instruments = {}
    else: 
        with open(instruments_file) as f: 
            instruments = json.loads(f.read())

    
    if kwargs.get('markov_values'):
        with open(kwargs['markov_values']) as f: 
            markov_values = json.loads(f.read())
    else: 
        markov_values = sample_markov.markov_values

    segments = [generate_segment(x, kwargs.get('beats_per_bar_range', range(3, 15)), bpm_range) for x in range(number_of_segments)]

    number_of_instruments_range = kwargs.get('number_of_song_instruments_range', range(1, random.randint(1, 3)))
    number_of_instruments = random.choice(number_of_instruments_range)
    if instruments: 
        song_instruments = [get_random_instrument(instruments, kwargs.get('instruments_range', range(0, 90))) for i in range(number_of_instruments)]
    else: 
        song_instruments = [random.choice(kwargs.get('instruments_range', range(0, 90))) for i in range(number_of_instruments)]
    print 'Song instruments are : ', song_instruments
        
    for i in range(len(segments)):

        create_segment(segments, i, kwargs, song_instruments, song_chord, markov_values, instruments = instruments)

    segments = shuffle_play_at(segments, kwargs.get('segment_shuffle_range', range(1, 3)))   
    print 'Generated segments are : '
    for segment in segments: 
        print 'Segment ', segment['track'], ' uses channels  : ',  [x['channel'] for x in segment['blocks']], ' has ', segment['play_at'], ' has  ', segment['number_of_beats_per_bar'], ' beats per bar and ', segment['number_of_bars'], ' bars and lasts ', segment['number_of_beats_per_bar'] * segment['number_of_bars'], ' playing at ', segment['bpm'], ' BPM and I repeat ', segment['number_repeats'], ' times. '

    base_block = template_utils.create_base_block()

    base_block['blocks'] = segments
    base_block['markov_values'] = markov_values


    gen_name = haikunator.Haikunator.haikunate()
    song_name = kwargs.get('generate_dir', 'generated/') + gen_name 


    mid = MIDIFile(100)
    
    entire_track = generator.handle_block(base_block, mid)
    mid = generator.generate_from_track(mid, entire_track, 100)    

#    mid = generator.generate(base_block)

    soundfont = kwargs.get('soundfont', 'soundfonts/FluidR3_GM.sf2')
    print 'Using soundfont : ', soundfont
    song_path = generator.write_mid(mid, song_name)
    if kwargs.get('get_mid'): 
        return song_path + '.mid'

    generator.to_wav(song_path, soundfont)
    success = subprocess.check_output(['lame', song_path + '.wav'])
    return song_path + '.mp3'

def generate_segment(track_no, beats_per_bar_range, bpm_range):
    segment =  {
        'name' : 'generated_track_no_' + str(track_no),
        'track' : track_no,
        'play_at' : [], 
        'number_of_beats_per_bar' : random.choice(beats_per_bar_range), 
        'bpm' : random.choice(bpm_range),
        'block_type' : 'complex', 
        'blocks' : []
    }
    return segment

def main(argv):
    if len(argv) > 2: 
        soundfont = argv[1]
        if len(argv) == 3: 
            generate_dir = argv[2]
    else: 
        soundfont = '../soundfonts/FluidR3_GM.sf2'
        generate_dir = '../http_generated/'
    generate_song(instruments_range = [1, 2, 4, 20, 21, 23, 24, 27, 34, 37, 39, 45, 47, 56, 68], segment_instruments_range = [1, 5, 8, 10, 11, 17, 24, 26, 28, 32, 33, 35, 36, 38, 39, 41, 45, 46, 52, 53, 54, 58, 68, 70, 72, 75, 76, 77, 78, 85], soundfont = soundfont, generate_dir = generate_dir)
#    generate_song(soundfont = soundfont, generate_dir = '../http_generated/')

#    generate_song(soundfont = soundfont, generate_dir = 'http_generated/')


if __name__ == '__main__' :
    main(sys.argv)



import random
import song_generator


#attributes_table is a table with keys that look like this: 
#   'key': 
#   {
#       'related_key' : 
#       [
#           {
#                'func' : get_key_values, 
#                'args' : ['some', 'args']
#                'prob' : 0.5,
#                'value' : 50,
#           },{
#                'prob' : 0.5,
#                'func' : get_key_values, 
#                'args' : ['some', 'args']
#                'value' : [
#                   {
#                       'prob' : 0.5, 
#                       'value' : 25
#                   }, {
#                       'prob' : 0.5, 
#                       'value' : 15
#                   }
#                ]
#           }
#       ]
#   }, {
#       'related_key_2':
#        ...
#
#How it then works is, we have the key that we want to make a decision for (called as decision key from now on). Then, we go key-by-key, find the decision key, randomly choose a value using the probability value ('prob') and then return the value for that key. 
#Values can be just a straight up value, but they can also be a list of dictionaries, each with their own probabilities and values, which can be further nested. 
#Alternatively, you may pass a function to the 'func' key, and args to the 'args' key. This function will then be called with the arguments to programatically generate the values. 


class DecisionMaker():
    def __init__(self, attributes, attributes_table):
        self.attributes_table = attributes_table
        for attribute in attributes: 
            setattr(self, attribute, attributes[attribute])

    def make_decision(self, attribute, related_keys = None): 
        attribute_table = [self.attributes_table[key] for key in self.attributes_table if key == attribute][0]
        attribute_value = self.get_table_result(attribute_table)
        

    def get_table_result(self, attribute_table):
        r, s = random.random(), 0
        for d in table:
            if type(d['value']) == list:
                d_value = self.get_table_result(d)
            else: 
                d_value = d['value']
            s += d['prob']
            if s >= r: 
                return d_value



import sys, json
from ninopianino import generator, template_utils

def seconds_to_min(seconds):
    return str(seconds/60) + ':' + str(seconds%60)

def instrument_play_time(block):
    return [seconds_to_min((block['program_number']-1)*4), seconds_to_min(block['program_number']*4)]

def generate_blocks():
    blocks = []
    for i in range(1, 90):
        new_block = {
            'program_number' : i, 
            'play_at' : [(i-1) * 8], 
            'number_of_bars' : 2, 
            'number_of_beats_per_bar' : 4,
            'default_accent' : 80,
        }
        blocks.append(new_block)
        play_time = instrument_play_time(new_block)
        print 'Program number ', i, 'between ', play_time[0], 'and', play_time[1] 
    return blocks

def main():
    base_block = template_utils.create_base_block()
    base_block['channel'] = 1
    base_block['root_note'] = 'C'
    base_block['scale'] = 'major'
    base_block['blocks'] = generate_blocks()
    base_block['low_end'] = 'C2'
    base_block['high_end'] = 'C3'
    mid = generator.generate(base_block)
    try: 
        soundfont = sys.argv[1]
    except: 
        import traceback
        traceback.print_exc()
        soundfont = 'soundfonts/FluidR3_GM.sf2' 
    print 'Testing ', soundfont

    song_path = generator.write_mid(mid, 'soundfont_test' )

    generator.to_wav(song_path, soundfont)

    instruments_json = [
        {
            'id' : 'nesho', 
            'friendly' : 'Nesho', 
            'base_volume' : 80, 
            'program_number' : i['program_number'], 
            'instrument_play_time' : instrument_play_time(i)[0] + ' - ' + instrument_play_time(i)[1]
        } for i in base_block['blocks']
    ]

    with open('instruments.json', 'w') as f: 
        f.write(json.dumps(instruments_json, indent = 4))




main()



import song_generator
from  markov_gen import markov_generator
import os, json


nino_dir = '/'.join(os.path.dirname(os.path.realpath(__file__)).split('/')[:-1])

import generator
generator.set_dir_write_note(nino_dir + '/trainer/generated_notes')


def gen_kwargs():
    kwargs = {
        #What the general scale for the shond should be - chosen randomly from this list. 
        'song_scale' : ['major', 'minor'],

        'block_same_note_range' : [0], 

        #How many segments the song has. 
        'number_of_segments_range' : [1],

        'chord_exp_var': 2,

    
        #The range of BPMs for each segment. Chooses randomly for each segment. 
        'bpm_range': [400], 

        #A range for beats per bar for each segment. Will choose randomly. 
        'beats_per_bar_range' : [4],

        #A range for how many chords each segment should have. Chooses randomly from this list. 
        'chords_per_segment_range': [1], 

        #A list containing the program numbers of instruments that should be used. The program will choose randomly from these. 
        'instruments_range' : [1, 2],

        #The number of instruments that will be active throughout the song. 
        'number_of_song_instruments_range' : [1],

        #The bias_same_note chances for the main instruments. It's a list that is randomly chosen from. 
        'main_instrument_bias_same_note' : [0],

        #The maximum pattern note length for the main instrument
        'pattern_note_max_len_range' : 1,

        #Each segment will have a different accent from the previous, determined by a random value from this list. 
        'accent_offset' : range(-5, 5),

        #A list from accents from which segments will receive their base accent. 
        'default_accent_range' : range(70, 85),

        #Volume to be added to percussions. 
        'percussion_accent_offset' : 10,

         #Number of extra instruments per segment. 
        'no_segment_instruments_range' : [0],

        #Range for the number of bars per segment. Will choose randomly from this list. 
        'number_segment_bars_range' : [32],

        #Accent range offset for instrument specific blocks. 
        'block_default_accent_range' :  range(-5, 5),

        #Chance for each instrument to follow a pattern for the duration of the segment. 
        'segment_instrument_pattern_chance' : 1.0,

        #Upper range for how long a pattern note can last. Should not be longer than the maximum amount of beats per bar for the instrument. 
        'pattern_note_len_range' :  1,
        
        #And the lower range for how long the note can last. Should not be less than 1. 
        'pattern_note_min_len_range' : 1,

        #The dir where the songs are saved. 
        'generate_dir' :  nino_dir + '/trainer/',

        #The directory for the soundfont. This is an example, and should be supplied for specific use cases. 
        'soundfont' :  nino_dir + '/soundfonts/FluidR3_GM.sf2',

        #The song generator will randomly repeat segments and then shuffle them. This is a range of the numbers of repeats for each segment. 
        'segment_shuffle_range' : [1],

        #We may want to have segments with few instruments and no drums. This is the percentage that there are drums if the number of instruments is below the defined treshold. 
        'segment_percussion_chance': 0.0,

        #If there's less than this many instruments in a segment, there's a chance (defined above) that there will be no percussion for that segment. 
        'skip_percussion_treshold' : 3,

        'get_mp3' : True,
        
        'dir_write_note' : nino_dir + '/trainer/generated_notes',
        'markov_values' : nino_dir + '/trainer/results.json',
    }
    return kwargs


def main():
    kwargs = gen_kwargs()
    song_generator.generate_song(**kwargs)

    markov_weight = float(raw_input('Enter a weight for the markov values. '))

    generated_notes = ''
    with open(nino_dir + '/trainer/generated_notes') as f: 
        generated_notes = '[' + f.read()[2:] + ']'

    generated_notes = json.loads(generated_notes)

    old_results = ''
    with open(nino_dir + '/trainer/results.json') as f: 
        old_results = json.loads(f.read())

    new_results = markov_generator.markov_from_values(old_results, generated_notes, 4, weight = markov_weight)
    with open(nino_dir + '/trainer/results.json', 'w') as f: 
        f.write(json.dumps(new_results))


if __name__ == '__main__' : 
    main()



import random, json, os


nino_dir = '/'.join(os.path.dirname(os.path.realpath(__file__)).split('/')[:-1])

def main():
    kwargs = {
        #What the general scale for the shond should be - chosen randomly from this list. 
        'song_scale' : ['major', 'minor'],

        #How many segments the song has. 
        'number_of_segments_range' : range(5, 8),
    
        #The range of BPMs for each segment. Chooses randomly for each segment. 
        'bpm_range': range(150, 500, 15), 

        #Probability that the bpm for a segment will be different from another one. 
        'bpm_change_prob': 0.4,

        #A range for beats per bar for each segment. Will choose randomly. 
        'beats_per_bar_range' : [3, 4, 4, 4, 4, 5],

        #A range for how many chords each segment should have. Chooses randomly from this list. 
        'chords_per_segment_range': range(2, 4), 

        #How many times each segment can repeat itself in succession. 
        'number_repeats_range' : range(1, 4),

        #File defining insturments
        'instruments_file' : nino_dir + '/ninopianino/instruments_sample.json',

        #The number of instruments that will be active throughout the song. 
        'number_of_song_instruments_range' : range(2, 4),

        #Number of main instruments per segment
        'number_main_instruments' : 1,

        #The bias_same_note chances for the main instruments. It's a list that is randomly chosen from. 
        'main_instrument_bias_same_note' : [0],

        #How much louder the main instrument should be.
        'main_default_accent_offset' : 20,

        #The maximum pattern note length for the main instrument
        'main_pattern_len_range' : 2, 

        #Each segment will have a different accent from the previous, determined by a random value from this list. 
        'accent_offset' : range(-5, 5),

        #A list from accents from which segments will receive their base accent. 
        'default_accent_range' : range(70, 85),

        #Volume to be added to percussions. 
        'percussion_accent_offset' : -20,

        #Number of extra instruments per segment. 
        'no_segment_instruments_range' : range(1, 3),

        #Range for the number of bars per segment. Will choose randomly from this list. 
        'number_segment_bars_range' : range(2, 5),

        #Accent range offset for instrument specific blocks. 
        'block_default_accent_range' :  range(-5, 5),

        #A list of values for bias_same_note for specific instruments. 
        'block_same_note_range' :  range(10, 40, 5),

        #Chance for each instrument to follow a pattern for the duration of the segment. 
        'segment_instrument_pattern_chance' : 0.7,

        #Upper range for how long a pattern note can last. Should not be longer than the maximum amount of beats per bar for the instrument. 
        'pattern_note_len_range' :  2,
        
        #And the lower range for how long the note can last. Should not be less than 1. 
        'pattern_note_min_len_range' : 1,

        #List for possible values for the bias_same_note for percussions per segment. 
        'percussion_bias_same_note' : range(30, 90, 5),

        #The same min and max values, but only for percussions. 
        'pattern_percussion_min_len_range': 2,
        'pattern_percussion_max_len_range': 4,

        #The dir where the songs are saved. 
        'generate_dir' :  nino_dir + '/normie_gen/',

        #The file from which we get markov values. 
#        'markov_values' : nino_dir + '/trainer/results2.json',
        'markov_values' : nino_dir + '/trainer/old_res.json', #pretty good


        #The directory for the soundfont. This is an example, and should be supplied for specific use cases.
        'soundfont' :  nino_dir + '/soundfonts/orchestra_sf/SGM-V2.01.sf2',

        #The song generator will randomly repeat segments and then shuffle them. This is a range of the numbers of repeats for each segment. 
        'segment_shuffle_range' : range(1, 5),

        #The script uses a specific function to generate a base default accent, but we may want to place an upper limit. 
        'max_base_default_accent' : 80,

        #This variable basically specifies how often the randomly generated chords will contain a chord vastly different from the generic chord. For instance, how commonly a chord progression for a song in C major to contain a chord in F minor. Higher values mean more generic chords. 
        'chord_exp_var': 4,

        #We may want to have segments with few instruments and no drums. This is the percentage that there are drums if the number of instruments is below the defined treshold. 
        'segment_percussion_chance': 0.85,

        #If there's less than this many instruments in a segment, there's a chance (defined above) that there will be no percussion for that segment. 
        'skip_percussion_treshold' : 3,

    }

    kwargs = json.dumps(kwargs)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(dir_path + '/sample_config.json', 'w') as f: 
        f.write(kwargs)

main()



import json

class Block:
    defaults = {
        'name' : 'unnamed',
        'track' : 1, 
        'channel' : 1, 
        'program_number' : 1, 
        'play_at' : [0], 
        'bpm' : 60, 
        'number_of_bars' : 1, 
        'number_of_beats_per_bar' : 4, 
        'bias_same_note' : 0, 
        'bias_separate_notes' : 0, 
        'pattern' : [], 
        'repeat' : 1, 
        'root_note' : 'C', 
        'scale' : 'major', 
        'base_notes' : [], 
        'low_end' : 'C1', 
        'high_end' : 'C6', 
        'default_accent' : 50, 
        'accents' : {}, 
        'notes_bias' : {}, 
        'block_type' : None, 
        'blocks' : []
    }
    
    
    hard_defaults = {
        'play_at' : [0], 
        'block_type' : None, 
    }
    def __init__(self, **kwargs):
        
        if kwargs.get('use_defaults', True): 
            for default in self.defaults: 
                if default not in kwargs: 
                    kwargs[default] = self.defaults[default]
                    
        for key in self.hard_defaults: 
            if key not in kwargs: 
                kwargs[key] = self.hard_defaults[key]
            
        for key in kwargs: 
            setattr(self, key, kwargs[key])

        print dir(self)

    def generate_dict_tree(self):
        if self.block_type: 
            temp_block = self
            blocks_temp = []
            for block in self.blocks:
                blocks_temp.append(block.generate_dict_tree())
            temp_block.blocks = blocks_temp
            return temp_block.__dict__
        else:
            d = self.__dict__
            return {x: d[x] for x in self.__dict__ if d[x]}


    def generate_json(self):
        dict_tree = self.generate_dict_tree()
        return json.dumps(dict_tree)
            
    def dump_json_to_file(self, file_path):
        with open(file_path, 'w') as f:
            t = self.generate_json()
            f.write(t)
            

def generate_percussion(block, hit_patterns = [], cymbal_patterns = []):
    percussions = Block(name = 'percussions', channel = 10, block_type = 'complex', number_of_bars = block.number_of_bars, number_of_beats_per_bar = block.number_of_beats_per_bar)
    
    for i in range(len(hit_patterns)):
        percussion_hits = Block(name = 'percussion_hits_' + str(i), pattern = hit_patterns[i], low_end = 'B0', high_end = 'G1', use_defaults = False)
        percussions.blocks = percussions.blocks + [percussion_hits]
        
    for i in range(len(cymbal_patterns)): 
        cymbal_hits = Block(name = 'cymbal_hits_' + str(i), pattern = cymbal_patterns[i], low_end = 'B0', high_end = 'G1', use_defaults = False)
        percussions.blocks= percussions.blocks + [cymbal_hits]
    return percussions



from midiutil.MidiFile import MIDIFile
import music_models, argparse, random, note_timing, copy, json, subprocess

#If this value is supplied, the Key class will write down all notes it generates in the file specified. 
dir_write_note = ''

def set_dir_write_note(new_dir):
    global dir_write_note
    dir_write_note = new_dir

def gen_notes_for_key(track, number_notes, root_note, scale, channel, duration = 1, bias_same_note = 0, low_end = 'A0', high_end = 'G#8', base_notes = [], notes_bias = {}, markov_values = None):
    k = music_models.Key(root_note, scale, base_notes, notes_bias, low_end, high_end, markov_values)
    notes = []
    prev_note = k.generate_note(None, 3)
    while number_notes>0:
        prev_note = k.generate_note(prev_note, 7, bias_same_note, dir_write_note = dir_write_note)
        notes.append(music_models.Note(channel = channel, track = track, note = prev_note, duration = duration, volume = 100)) 
        number_notes -= 1 
    return notes

def calculate_number_of_notes(block):
    if block.get('number_of_notes'): return block['number_of_notes']
    elif block.get('number_of_bars'): return block['number_of_bars'] * block['number_of_beats_per_bar']
    else: raise Exception('There was an error calculating the number of notes in block ', block.get('name'))


def generate_generic_notes(b):
    number_of_notes = calculate_number_of_notes(b)
    gen_notes_kwargs = {'track' : b['track'], 'channel' : b.get('channel', 1), 'number_notes' : number_of_notes, 'root_note' : b.get('root_note', 'A'), 'scale' : b.get('scale', 'minor'), 'bias_same_note' : b.get('bias_same_note'), 'high_end' : b.get('high_end', 'G#7'), 'low_end' : b.get('low_end', 'C1'), 'base_notes': b.get('base_notes'), 'notes_bias': b.get('notes_bias', {}), 'markov_values' : b.get('markov_values')}
    generic_notes = gen_notes_for_key(**gen_notes_kwargs)
    return generic_notes


def group_generic_notes(b, generic_notes, starting_point):
    ungrouped_notes = copy.deepcopy(generic_notes)
    if b.get('accents') : 
        accents = {int(x):b['accents'][x] for x in b['accents']}
    else : accents = {}
    grouped_notes_kwargs = {'notes' : ungrouped_notes, 'no_beats' : b.get('number_of_beats_per_bar'), 'time_signature' : b.get('time_signature'), 'bias_separate_notes' : b.get('bias_separate_notes'), 'accents' : accents, 'start_at' : starting_point, 'pattern' : b.get('pattern', []), 'default_accent':b.get('default_accent', 50)}
    grouped_notes = note_timing.group_notes_for_time_signature(**grouped_notes_kwargs)
    return grouped_notes    


def handle_block(b, mid):
    if b.get('repeat', 1) > 1:
        b['play_at'] += [i * b.get('number_of_beats_per_bar', 1) * b.get('number_of_bars') for i in range(1, b['repeat']+1)]
    if b.get('block_type') == 'complex' :
        mid.addTempo(b['track'], b['play_at'][0], b['bpm']) 
        complex_track = []
        for block in b['blocks']: 
            if not block: continue 
            for key in b:
                if key not in block.keys() + ['blocks', 'block_type', 'play_at', 'repeat', 'number_of_blocks']:
                    block[key] = b[key]
            complex_track += handle_block(block, mid)

        entire_track = []
        
        for starting_point in b['play_at']:
            temp_track = copy.deepcopy(complex_track)
            for bar in temp_track: 
                for note in bar.notes:
                    note.time += starting_point
            entire_track += temp_track
    else:
        entire_track = []
        generic_notes = generate_generic_notes(b)       
            
        for starting_point in b['play_at']:
                mid.addProgramChange(b['track'], b.get('channel', 2) - 1, starting_point, b.get('program_number', 0)) 
#                print 'Adding program change for track : ', b['track'], ' on channel : ', b.get('channel') - 1
                grouped_notes = group_generic_notes(b, generic_notes, starting_point)
                entire_track += grouped_notes
#        print 'notes for ', b['name'], ':', [[t.time for t in x.notes] for x in entire_track], '\n\n'
    return entire_track





def main():

    parser = argparse.ArgumentParser(description="Basic arguments")
    parser.add_argument('--use_soundfont', help = 'Soundfont to use for creating a wav file. If not specified, will just create a mid file', default = '')
    parser.add_argument('--output', help = 'Path for output midi file. ', default = 'output')
    parser.add_argument('--input', help = 'Path for input JSON file. ', default = 'input.JSON')
    parser.add_argument('--no_tracks', help = 'Number of tracks. Default is 100. ', default = 100)

    args = parser.parse_args()
    output = args.output
    blocks = json.loads(open(args.input, 'r').read())
    
    no_tracks = args.no_tracks
    
    mid = generate(blocks, no_tracks)
    write_mid(mid, args.output, args.use_soundfont)
    
    
def generate(blocks, no_tracks = 100):
    mid = MIDIFile(no_tracks)

    entire_track = handle_block(blocks, mid)
    return generate_from_track(mid, entire_track, no_tracks)    

def generate_from_track(mid, entire_track, no_tracks = 100):
    for bar in entire_track: 
        for note in bar.notes:
            mid.addNote(*note.get_values())                

    return mid
   
def write_mid(mid, output):
    binfile = open(output + '.mid', 'w+b')
    
    mid.writeFile(binfile)
    binfile.close()
    return output

def to_wav(song_path, soundfont):
    command = ['fluidsynth', '-F', song_path + '.wav', soundfont, song_path + '.mid']
    subprocess.call(command)


if __name__ == '__main__' : 
    main()
